/* Topology Builder        Bruce D. Ray
   IUPUI Physics Dept. NMR Center
   402 N. Blackford St.
   Indianapolis, IN   46202
   USA

        Functions to read and use gromacs force field data
        gromacs_use.c
        Portions of this work are copied from the files read_amber.c,
        and the do_an_amber function in the file topolbuild.c
*/

#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include "gromacs_use.h"

extern char mess[MAXCHAR];
extern char ambername[MAXCHAR];
extern char mycorr_name[MAXCHAR];
extern char line[2*MAXCHAR];

/* code to use a gromacs atoms force field, including united atoms model
   options
   parameters are:
        ante_name        directory of antechamber atom type data files
        leap_name        directory of leap amber and gaff parm files
        num_atoms        number of atoms in molecule
        num_bonds        number of bonds in molecule
        amberff          specific name of force field to use
        angle_max        number of angles in molecule
        tors_max         number of torsions in molecule
        ua_model         flag for use of united atoms model
        residue          residue name or NULLCHAR if do not have
                         common residue.
        by_corr          flag for set final atom types by correlation
                         or by second atom type definition file.
*/

void do_gmx_FFs(char *ante_name, char *leap_name, int num_atoms,
                int num_bonds, char *amberff, int angle_max,
                int tors_max, int ua_model, char *residue, int by_corr)
{
     int num, num2, count1, count2, i;
     int FFmol, FFbond, FFangle, FFtors, FFimpro, FFvdw, FFarom;

/* get a first level gromacs atom types assignment  Needed in order to
   properly sort out the various parameter differences for the same
   standard gromacs atom type in similar situations, but in different
   amino acid residues. Use a modified gaff atom types definition table
   for this step.
*/

     strcpy(ambername, ante_name);
     strcat(ambername, "/ATOMTYPE_GMX1.DEF");

     read_at_types(&num, &num2, (2 * MAXCHAR), MAXCHAR,
                   0, ambername);                   /* first count lines */

     fprintf(stdlog,
             "\nIn %s there are %d WILDATOM lines and %d ATD lines\n",
             ambername, num2, num);
     count1 = num;
     count2 = num2;
     read_at_types(&num, &num2, count1, count2, 1,
                   ambername);                      /* now read them */

     init_atom_types(num_atoms, num_bonds);
     bond_info(num_atoms, num_bonds);

     judge_atom_types(count1, count2, num_atoms, num_bonds, a_ff_type);
     adjustments(num_atoms, num_bonds, a_ff_type);
     adjust_type_cp(num_atoms, num_bonds, a_ff_type);
     check_type_errors(num_atoms, num_bonds, a_ff_type);

     free_atom_types(num_atoms);
     release_types(count1, count2);

/* now have the first level of gromacs type assignments and can work
   on the force field entries and assignments
*/

     strcpy(ambername, leap_name);
     strcat(ambername, "/ff");
     strcat(ambername, amberff);
     strcat(ambername, ".dat");

/* This is my own forcefield format and has the counts of each type of
   before entry as part of the file.  Therefore I don't have to count
   anything allocating memory
*/

     read_ffgmx1(ambername, &FFmol, &FFbond, &FFangle, &FFtors,
                 &FFimpro, &FFvdw, &FFarom);

     fprintf(stdlog,
             "\n\tFile has\n\t\t%d atom records\n\t\t%d bond records\n",
             FFmol, FFbond);
     fprintf(stdlog, "\t\t%d angle records\n", FFangle);
     fprintf(stdlog,
             "\t\t%d torsion records\n\t\t%d impropers records\n",
             FFtors, FFimpro);
     fprintf(stdlog, "\t\t%d van der Waals records\n", FFvdw);

/* I arranged my format so that I can use the same matching routines that
   I use with amber force field parameters
*/

     set_amber_bondFF(num_bonds, FFbond);
     set_amber_angleFF(angle_max, FFangle);
     set_gromacs_torsFF(tors_max, FFtors);
     if(improper_num)
          set_amber_improFF(improper_num, FFimpro);

     if(ua_model)                            /* convert to united atoms model */
          united_atoms(num_atoms, residue, FFarom);     /* if called for */

/* either use correlations between types, or use another atom type
   definition to get to the gromacs atom types.  Which is used depends
   on how the particular gromacs force field can best be set up.
*/
     if(by_corr) {
          strcpy(mycorr_name, ante_name);
          strcat(mycorr_name, "/");
          strcat(mycorr_name, amberff);
          strcat(mycorr_name, ".corr");

/* Now get the gromacs atom type correlation names to assign the
   original gromacs atom types.  Again, this is my own format and
   includes the needed counts in the file so that I don't have to
   go through a count cycle before allocating memory.
*/

          read_gmx_corr(mycorr_name, num_atoms);
     }
     else {
          strcpy(mycorr_name, ante_name);
          strcat(mycorr_name, "/ATOMTYPE_");
          strcat(mycorr_name, amberff);
          strcat(mycorr_name, ".DEF");

          read_at_types(&num, &num2, (2 * MAXCHAR), MAXCHAR,
                        0, mycorr_name);            /* first count lines */

          fprintf(stdlog,
                  "\nIn %s there are %d WILDATOM lines and %d ATD lines\n",
                  mycorr_name, num2, num);
          count1 = num;
          count2 = num2;
          read_at_types(&num, &num2, count1, count2, 1,
                        mycorr_name);               /* now read them */

          init_atom_types(num_atoms, num_bonds);
          bond_info(num_atoms, num_bonds);

          judge_atom_types(count1, count2, num_atoms, num_bonds, a_ff_type);

/* not a gaff type atom type definition file.  gaff atom type
   corrections won't work here and aren't done.
*/
          free_atom_types(num_atoms);
          release_types(count1, count2);
     }

     if(FFvdw > 2)
          set_amber_vdwFF(num_atoms, FFvdw);
     set_amber_atomFF(num_atoms, FFmol);

     free_me(gmx_tors_resid, FARGS);
     free_me(gmx_tors_num, FARGS);
     free_amberFF(FFmol, FFbond, FFangle, FFtors, FFimpro, FFvdw);

     if(FFarom)
          free_me(gmx_aroms, FARGS);

     free_me(macro_types, FARGS);
     free_me(macro_name, FARGS);
     free_me(macro_nmbr, FARGS);

     return;
}

/* read correlations between gromacs atom types and the intermediate
   step atom types, and assign corrected atom types.
   parameters are:
        mycorr_name               correlation file name
        num_atoms                 number of atoms
*/
void read_gmx_corr(char *mycorr_name, int num_atoms)
{
     int i, j, correls;
     FILE *atcorrelp;
     char **correlin, **correlto;

     if((atcorrelp = fopen(mycorr_name, "r")) == NULL) {
          sprintf(mess,
                  "Cannot open atom type correlation file %s\n",
                  mycorr_name);
          my_fatal(FARGS, mess);
          exit(1);
     }

     correls = 0;
     while(fgets(line, MAXCHAR, atcorrelp) != NULL) {
          if(!strncmp(line, "correls", 7)) {
               sscanf(&line[8], "%d", &correls);
               break;
          }
     }

     if(!correls) {
          sprintf(mess,
                  "Atom type correlation file %s lacks count data\n",
                  mycorr_name);
          my_fatal(FARGS, mess);
          exit(1);
     }

     correlin = (char **)mat_alloc(correls, 6, FARGS);
     correlto  = (char **)mat_alloc(correls, 9, FARGS);

     for(i = 0; i < correls; i++) {
          if(fgets(line, MAXCHAR, atcorrelp) == NULL) {
               sprintf(mess,
                       "Premature end to atom type correlation file %s\n",
                       mycorr_name);
               my_fatal(FARGS, mess);
               exit(1);
          }
          sscanf(line, "%s %s", correlin[i], correlto[i]);
     }

     fclose(atcorrelp);

     for(i = 0; i < num_atoms; i++) {
          for(j = 0; j < correls; j++){
               if(!strcmp(a_ff_type[i], correlin[j])) {
                    strcpy(a_ff_type[i], correlto[j]);
                    break;
               }
          }
     }

     for(i = 0; i < correls; i++) {
          free_me(correlin[i], FARGS);
          free_me(correlto[i], FARGS);
     }
     free_me(correlin, FARGS);
     free_me(correlto, FARGS);

     return;
}

/* convert C to united atoms model
   parameter are:
        num_atoms          number of atoms
        resid              residue name or NULLCHAR if do not have
                           common residue.
*/
void united_atoms(int num_atoms, char *resid, int nmarom)
{
     int i, j, k, l, any, count;
     char mixer[6], *c, *d;

     any = 0;
     for(i = 0; i < num_atoms; i++) {
          if(atom_atno[i] != 6) continue;      /* Only looking for C */
          if(!bond_count[i]) continue;         /* Only C with bonds to
                                                  something else */

          if(nmarom) {
               strncpy(mixer, atom_residue[i], 5);
               mixer[5] = NULLCHAR;
               if((c = strpbrk(mixer, "0123456789")) != NULL) {
                    k = strlen(mixer);
                    if((c != &mixer[0]) &&
                       ((d = strpbrk(&mixer[k-1], "0123456789"))
                        != NULL))
                              *c = NULLCHAR;
               }                      /* end if((c = strpbrk  etc. */

               k = 0;
               for(j = 0; j < nmarom; j++) {           /* Don't check
                                                          listed
                                                          carbons */
                    if((gmx_aroms[j][1][0] == '*') ||
                       (!strcmp(gmx_aroms[j][1], mixer))) 
                              if(!strcmp(a_ff_type[i],
                                 gmx_aroms[j][0])) {
                                        k = 1;
                                        break;
                              }       /* end if string */
               }                      /* end for(j = 0; j < nmarom; j++) */
               if(k) continue;
          }                           /* end if(nmarom) */

          for(j = 0; j < bond_count[i]; j++) {
               if(atom_atno[(connect[i][j])] != 1)
                    continue;   /* if not H, skip it */
/* add charge of the H to the C */
               atom_charge[i] += atom_charge[(connect[i][j])];
/* add mass of the H to the C */
               atom_mass[i] += atom_mass[(connect[i][j])];
/* remove mass from the H */
               atom_mass[(connect[i][j])] = 0.0;
               any = 1;               /* yes, I did one */
          }                           /* end for(j = 0; j < bond_count[i]; j++) */
     }                                /* end for(i = 0; i < num_atoms; i++) */

     if(!any) return;                 /* Only need to do more if changes made */

     for(i = 0; i < num_atoms; i++) {
          atom_order[i] = -1;
          back_order[i] = -1;
     }

     count = 0;
     if(resid[0] != NULLCHAR) {
          for(i = 0; i < num_atoms; i++) {
               if(atom_mass[i] == 0.0) {
                    atom_order[i] = -1;
                    back_order[i] = -1;
                    continue;
               }
               atom_order[i] = i;
               back_order[i] = count;
               count++;
          }                /* end for(i = 0; i < num_atoms; i++) */

          true_atoms = count;

          return;
     }                     /* end if(resid[0] != NULLCHAR) */

     j = 0;
     k = 0;

     while(j < num_atoms) {
          k++;
          for(i = 0; i < num_atoms; i++) {
               if(back_order[i] > -1)
                    continue;                 /* already set? */
               if(k == atom_segno[i]) {
                    if(atom_mass[i] == 0.0) {
                         atom_order[j] = -1;
                         back_order[i] = -1;
                         j++;
                         continue;
                    }
                    atom_order[j] = i;
                    back_order[i] = count;
                    j++;
                    count++;
/* set attached H's next to parent */
                    if((bond_count[i] > 0) && (atom_atno[i] > 1))
                         for(l = 0; l < bond_count[i]; l++) {
                              if(atom_mass[(connect[i][l])] == 0.0)
                                   continue;
                              if(atom_atno[(connect[i][l])] == 1) {
                                   atom_order[j] = connect[i][l];
                                   back_order[(connect[i][l])] = count;
                                   j++;
                                   count++;
                              }  /* end if(atom_atno[(connect[i][l])] == 1) */
                         }       /* end for(l = 0; l < bond_count[i]; l++) */
                                 /* and end if((bond_count[i] > 0) &&
                                    (atom_atno[i] > 1)) */
               }                 /* end if(k == atom_segno[i]) */
          }                      /* end for(i = 0; i < num_atoms; i++) */
     }                           /* while(j < num_atoms) */

     true_atoms = count;

     return;
}

/* read a gromacs intermediate atom types force field
   parameters are:
     
*/
void read_ffgmx1(char *ffname, int *nmmol, int *nmbond, int *nmangle,
                 int *nmtors, int *nmimpro, int *nmvdw, int *nmarom)
{
     FILE *ffptr;
     char myfiles[MAXCHAR], my_type[8], *c, *d;
     int i, j, k, countin, vdwtype;
     double dum1, dum2;
     int set_same_tors = 0;

     strcpy(myfiles, ffname);

     if((ffptr = fopen(myfiles, "r")) == NULL) {
          sprintf(mess,
                  "Cannot open atomic parameters file %s\n",
                  myfiles);
          my_fatal(FARGS, mess);
          exit(1);
     }

/* get macro names parameters */
     countin = 0;
     while(fgets(line, MAXCHAR, ffptr) != NULL) {
          if(!strncmp(line, "equivs", 6)) {
               sscanf(&line[7], "%d", &countin);
               break;
          }
     }

     if(!countin) {
          sprintf(mess,
                  "Force field file %s lacks macro names count data\n",
                  myfiles);
          my_fatal(FARGS, mess);
          exit(1);
     }

     macro_name = (char **)mat_alloc(countin, 4, FARGS);
     macro_nmbr = (int *)allocator(countin, sizeof(int), FARGS);
     macro_types = (char ***)allocator(countin, sizeof(char**), FARGS);
     num_macro = countin;
     i = 0;
/* use of while instead of for to permit comments and blank lines */
     while(i < countin) {
          if(fgets(line, MAXCHAR, ffptr) == NULL) {
               sprintf(mess,
                       "Premature end to force field file %s\n",
                       myfiles);
               my_fatal(FARGS, mess);
               exit(1);
          }
          if((c = strchr(line, '\n')) != NULL)    /* strip returns */
               *c = NULLCHAR;
          if((c = strchr(line, ';')) != NULL)
               *c = NULLCHAR;
          if((j = strspn(line, " \t")) == strlen(line))
               continue;                    /* drop blank lines */
          if(!strlen(line))                 /* drop empty strings */
               continue;                    /* as well */
          strncpy(macro_name[i], line, 3);  /* all macro names are */
          macro_name[i][3] = NULLCHAR;      /* 3 characters */
          sscanf(&line[5], "%d", &macro_nmbr[i]);
          macro_types[i] = (char **)mat_alloc(macro_nmbr[i], 6, FARGS);

/* list of atom types without comments or blanks follows */
          for(j = 0; j < macro_nmbr[i]; j++) {
               if(fgets(line, MAXCHAR, ffptr) == NULL) {
                    sprintf(mess,
                            "Premature end to force field file %s\n",
                            myfiles);
                    my_fatal(FARGS, mess);
                    exit(1);
               }
               if((c = strchr(line, '\n')) != NULL)
                    *c = NULLCHAR;              /* strip returns */
               if((c = strchr(line, ';')) != NULL)
                    *c = NULLCHAR;
               c = get_name(line, macro_types[i][j], 0);
          }          /* end for(j = 0; j < macro_nmbr[i]; j++) */
          i++;
     }               /* end while(i < countin) */

/* get atoms parameters including van der Waals */
     countin = 0;
     vdwtype = 0;
     while(fgets(line, MAXCHAR, ffptr) != NULL) {
          if(!strncmp(line, "atoms", 5)) {
               sscanf(&line[6], "%d %d", &countin, &vdwtype);
               break;
          }
     }

     if(!countin) {
          sprintf(mess,
                  "Force field file %s lacks atom types count data\n",
                  myfiles);
          my_fatal(FARGS, mess);
          exit(1);
     }

     ffatom_name = (char **)mat_alloc(countin, 6, FARGS);
     ffatom_mass = (float *)allocator(countin, sizeof(float), FARGS);
     ffatom_pol = (float *)allocator(countin, sizeof(float), FARGS);
     vdw_atom_name = (char **)mat_alloc(countin, 6, FARGS);
     vdw_atom_radius = (float *)allocator(countin, sizeof(float), FARGS);
     vdw_atom_pot = (float *)allocator(countin, sizeof(float), FARGS);
     *nmmol = countin;
     *nmvdw = countin;

     i = 0;
/* use of while instead of for to permit comments and blank lines */
     while(i < countin) {
          if(fgets(line, MAXCHAR, ffptr) == NULL) {
               sprintf(mess,
                       "Premature end to force field file %s\n",
                       myfiles);
               my_fatal(FARGS, mess);
               exit(1);
          }
          if((c = strchr(line, '\n')) != NULL)
               *c = NULLCHAR;                     /* strip returns */
          if((c = strchr(line, ';')) != NULL)
               *c = NULLCHAR;
          if((j = strspn(line, " \t")) == strlen(line))
               continue;                          /* drop blank lines */
          if(!strlen(line))
               continue;                /* drop empty strings as well */
          c = get_name(line, ffatom_name[i], 0);
          strcpy(vdw_atom_name[i], ffatom_name[i]);
          vdw_atom_radius[i] = 99999.0;
          vdw_atom_pot[i] = 99999.0;
          ffatom_mass[i] = 99999.0;
          ffatom_pol[i] = 99999.0;
          sscanf(c, "%f %f %lf %lf", &ffatom_mass[i], &ffatom_pol[i],
                 &dum1, &dum2);
          if((vdwtype == 6) && (dum2 > 0.0)) {
               vdw_atom_radius[i] = CubeRoot((fast_InvSqrt((dum1/dum2))));
               vdw_atom_pot[i] = (dum1 * dum1)/4/dum2;
          }
          else {
               vdw_atom_radius[i] = dum1;
               vdw_atom_pot[i] = dum2;
          }
          i++;
     }

/* get bonds parameters */
    countin = 0;
     while(fgets(line, MAXCHAR, ffptr) != NULL) {
          if(!strncmp(line, "bonds", 5)) {
               sscanf(&line[6], "%d", &countin);
               break;
          }
     }

     if(!countin) {
          sprintf(mess,
                  "Force field file %s lacks bonds count data\n",
                  myfiles);
          my_fatal(FARGS, mess);
          exit(1);
     }

     bond_name1 = (char **)mat_alloc(countin, 6, FARGS);
     bond_name2 = (char **)mat_alloc(countin, 6, FARGS);
     bond_force = (float *)allocator(countin, sizeof(float), FARGS);
     bond_beta = (float *)allocator(countin, sizeof(float), FARGS);
     bond_length = (float *)allocator(countin, sizeof(float), FARGS);
     FF_bond_type = (short *)allocator(countin, sizeof(short), FARGS);
     *nmbond = countin;

     i = 0;
/* use of while instead of for to permit comments and blank lines */
     while(i < countin) {
          if(fgets(line, MAXCHAR, ffptr) == NULL) {
               sprintf(mess,
                       "Premature end to force field file %s\n",
                       myfiles);
               my_fatal(FARGS, mess);
               exit(1);
          }
          if((c = strchr(line, '\n')) != NULL)
               *c = NULLCHAR;                    /* strip returns */
          if((c = strchr(line, ';')) != NULL)
               *c = NULLCHAR;
          if((j = strspn(line, " \t")) == strlen(line))
               continue;                         /* drop blank lines */
          if(!strlen(line))
               continue;               /* drop empty strings as well */
          bond_force[i] = -1.0;
          bond_length[i] = 99999.0;
          FF_bond_type[i] = 0;
          c = get_name(line, bond_name1[i], 0);
          c = get_name(c, bond_name2[i], 0);
          c = get_name(c, my_type, 0);
          if((d = strchr(my_type, ' ')) != NULL)
               *d = NULLCHAR;
          k = strspn(my_type, "0123456789");
          if(k != strlen(my_type)) {
               sprintf(mess,
                 "Illegal characters in bond type, force field file %s\n",
                       myfiles);
               my_fatal(FARGS, mess);
               exit(1);
          }
          FF_bond_type[i] = atoi(my_type);
          switch(FF_bond_type[i]) {
               case 5:  break;

               case 1:
               case 2:
               case 6:
               case 8:
               case 9:
               case 7:  sscanf(c, "%f %f", &bond_length[i],
                               &bond_force[i]);
                        break;

               case 3:
               case 4:  sscanf(c, "%f %f %f", &bond_length[i],
                               &bond_force[i], &bond_beta[i]);
                        break;

               default: sprintf(mess,
                          "Illegal bond type in parameters, force field file %s\n",
                          myfiles);
                        my_fatal(FARGS, mess);
                        exit(1);
          }
          i++;
     }

/* get angles parameters */
     countin = 0;
     while(fgets(line, MAXCHAR, ffptr) != NULL) {
          if(!strncmp(line, "angles", 6)) {
               sscanf(&line[7], "%d", &countin);
               break;
          }
     }

     if(!countin) {
          sprintf(mess,
                  "Force field file %s lacks angles count data\n",
                  myfiles);
          my_fatal(FARGS, mess);
          exit(1);
     }

     angle_name1 = (char **)mat_alloc(countin, 6, FARGS);
     angle_name2 = (char **)mat_alloc(countin, 6, FARGS);
     angle_name3 = (char **)mat_alloc(countin, 6, FARGS);
     angle_quartic = (float **)allocator(5, sizeof(float *), FARGS);
     for(i = 0; i < 5; i++)
          angle_quartic[i] = (float *)allocator(countin,
                                                sizeof(float),
                                                FARGS);
     angle_force = angle_quartic[0];
     angle_angle = (float *)allocator(countin, sizeof(float),
                                      FARGS);
     FF_angle_type = (short *)allocator(countin, sizeof(short),
                                        FARGS);
     *nmangle = countin;

     i = 0;
/* use of while instead of for to permit comments and blank lines */
     while(i < countin) {
          if(fgets(line, MAXCHAR, ffptr) == NULL) {
               sprintf(mess,
                       "Premature end to force field file %s\n",
                       myfiles);
               my_fatal(FARGS, mess);
               exit(1);
          }
          if((c = strchr(line, '\n')) != NULL)
               *c = NULLCHAR;                   /* strip returns */
          if((c = strchr(line, ';')) != NULL)
               *c = NULLCHAR;
          if((j = strspn(line, " \t")) == strlen(line))
               continue;                        /* drop blank lines */
          if(!strlen(line))
               continue;              /* drop empty strings as well */
          angle_force[i] = -1.0;
          angle_angle[i] = 99999.0;
          FF_angle_type[i] = 0;
          c = get_name(line, angle_name1[i], 0);
          c = get_name(c, angle_name2[i], 0);
          c = get_name(c, angle_name3[i], 0);
          c = get_name(c, my_type, 0);
          if((d = strchr(my_type, ' ')) != NULL)
               *d = NULLCHAR;
          k = strspn(my_type, "0123456789");
          if(k != strlen(my_type)) {
               sprintf(mess,
                  "Illegal characters in angle type, force field file %s\n",
                       myfiles);
               my_fatal(FARGS, mess);
               exit(1);
          }
          FF_angle_type[i] = atoi(my_type);
          switch(FF_angle_type[i]) {
               case 1:
               case 2:  sscanf(c, "%f %f", &angle_angle[i],
                               &angle_force[i]);
                        break;

               case 3:  sscanf(c, "%f %f %f", &angle_quartic[0][i],
                               &angle_quartic[1][i],
                               &angle_quartic[2][i]);
                        break;

               case 4:  sscanf(c, "%f %f %f %f", &angle_quartic[0][i],
                               &angle_quartic[1][i], &angle_quartic[2][i],
                               &angle_quartic[3][i]);
                        break;

               case 5:  sscanf(c, "%f %f %f %f", &angle_angle[i],
                               &angle_quartic[0][i], &angle_quartic[1][i],
                               &angle_quartic[2][i]);
                        break;

               case 6:  sscanf(c, "%f %f %f %f %f %f", &angle_angle[i],
                               &angle_quartic[0][i], &angle_quartic[1][i],
                               &angle_quartic[2][i], &angle_quartic[3][i],
                               &angle_quartic[4][i]);
                        break;

               case 8:  sscanf(c, "%f %f", &angle_quartic[0][i],
                               &angle_quartic[1][i]);
                        break;

               default: sprintf(mess,
                          "Illegal angle type in parameters, force field file %s\n",
                                 myfiles);
                        my_fatal(FARGS, mess);
                        exit(1);
          }
          i++;
     }

/* get regular dihedrals parameters */
     countin = 0;
     while(fgets(line, MAXCHAR, ffptr) != NULL) {
          if(!strncmp(line, "torsions", 8)) {
               sscanf(&line[9], "%d", &countin);
               break;
          }
     }

     if(!countin) {
          sprintf(mess,
                  "Force field file %s lacks torsions count data\n",
                  myfiles);
          my_fatal(FARGS, mess);
          exit(1);
     }

     tors_name1 = (char **)mat_alloc(countin, 6, FARGS);
     tors_name2 = (char **)mat_alloc(countin, 6, FARGS);
     tors_name3 = (char **)mat_alloc(countin, 6, FARGS);
     tors_name4 = (char **)mat_alloc(countin, 6, FARGS);
     tors_RB = (float **)allocator(6, sizeof(float *), FARGS);
     for(i = 0; i < 6; i++)
          tors_RB[i] = (float *)allocator(countin, sizeof(float),
                                          FARGS);
     tors_mult = (int *)tors_RB[0];
     tors_force = tors_RB[1];
     tors_phase = tors_RB[2];
     tors_term = tors_RB[3];
     tors_more = (int *)tors_RB[4];
     FF_tors_type = (short *)allocator(countin, sizeof(short),
                                       FARGS);
     gmx_tors_resid = (char ***)allocator(countin, sizeof(char **),
                                          FARGS);
     gmx_tors_num = (int *)allocator(countin, sizeof(int), FARGS);
     *nmtors = countin;
     set_same_tors = 0;

     i = 0;
/* use of while instead of for to permit comments and blank lines */
     while(i < countin) {
          if(fgets(line, MAXCHAR, ffptr) == NULL) {
               sprintf(mess,
                       "Premature end to force field file %s\n",
                       myfiles);
               my_fatal(FARGS, mess);
               exit(1);
          }
          if((c = strchr(line, '\n')) != NULL)
               *c = NULLCHAR;                      /* strip returns */
          if((c = strchr(line, ';')) != NULL)
               *c = NULLCHAR;
          if((j = strspn(line, " \t")) == strlen(line))
               continue;                        /* drop blank lines */
          if(!strlen(line))
               continue;               /* drop empty strings as well */
          tors_more[i] = 0;
          gmx_tors_resid[i] = NULL;
          gmx_tors_num[i] = 0;
          tors_mult[i] = 9999;
          tors_force[i] = -1.0;
          tors_phase[i] = 99999.0;
          tors_term[i] = 1.0;
          c = get_name(line, tors_name1[i], 0);
          c = get_name(c, tors_name2[i], 0);
          c = get_name(c, tors_name3[i], 0);
          c = get_name(c, tors_name4[i], 0);
          c = get_name(c, my_type, 0);
          if((d = strchr(my_type, ' ')) != NULL)
               *d = NULLCHAR;
          k = strspn(my_type, "0123456789");
          if(k != strlen(my_type)) {
               sprintf(mess,
                  "Illegal characters in torsion type, force field file %s\n",
                       myfiles);
               my_fatal(FARGS, mess);
               exit(1);
          }
          FF_tors_type[i] = atoi(my_type);
          switch(FF_tors_type[i]) {
               case 1:  sscanf(c, "%f %f %d %d", &tors_phase[i],
                               &tors_force[i], &tors_mult[i],
                               &gmx_tors_num[i]); 
                        if(gmx_tors_num[i] > 10) {
                             sprintf(mess,
                               "Bad torsions residue count in force field file %s\n",
                                     myfiles);
                             my_fatal(FARGS, mess);
                             exit(1);
                        }
                        tors_term[i] = (float)abs(tors_mult[i]);
                        if(gmx_tors_num[i]) {
                             if(fgets(line, MAXCHAR, ffptr) == NULL) {
                                  sprintf(mess,
                                    "Premature end to force field file %s\n",
                                          myfiles);
                                  my_fatal(FARGS, mess);
                                  exit(1);
                             }
                             if((c = strchr(line, '\n')) != NULL)
                                  *c = NULLCHAR;     /* strip returns */
                             if((c = strchr(line, ';')) != NULL)
                                  *c = NULLCHAR;
                             gmx_tors_resid[i] = (char **)mat_alloc(gmx_tors_num[i],
                                                                    6, FARGS);
                             c = &line[0];
                             for(j = 0; j < gmx_tors_num[i]; j++)
                                  c = get_name(c, &gmx_tors_resid[i][j][0], 0);
                        }

                        while(set_same_tors) {
                             if(gmx_tors_num[i] != gmx_tors_num[(i - 1)]) {
                                  set_same_tors = 0;
                                  break;
                             }
                             if(!gmx_tors_num[i])
                                  break;
                             for(j = 0; j < gmx_tors_num[i]; j++) 
                                  if(strcmp(gmx_tors_resid[i][j],
                                            gmx_tors_resid[(i - 1)][j])) {
                                       set_same_tors = 0;
                                       break;
                                  }
                             break;      /* needs a break when using
                                               a while as an if */
                        }                /* end while(set_same_tors) */
                        if(set_same_tors) {
                             if(!strcmp(tors_name1[i],
                                        tors_name1[(i - 1)]) &&
                                !strcmp(tors_name2[i],
                                        tors_name2[(i - 1)]) &&
                                !strcmp(tors_name3[i],
                                        tors_name3[(i - 1)]) &&
                                !strcmp(tors_name4[i],
                                        tors_name4[(i - 1)])) 
                                     flag_more_tors(i);
                             else set_same_tors = 0;
                        }

                        if(tors_mult[i] < 0) set_same_tors = 1;
                        else set_same_tors = 0;
                        break;

               case 3:  sscanf(c, "%f %f %f %f %f %f", &tors_RB[0][i],
                               &tors_RB[1][i], &tors_RB[2][i],
                               &tors_RB[3][i], &tors_RB[4][i],
                               &tors_RB[5][i]);
                        set_same_tors = 0;
                        break;

               case 5:  sscanf(c, "%f %f %f %f", &tors_RB[0][i],
                               &tors_RB[1][i], &tors_RB[2][i],
                               &tors_RB[3][i]);
                        set_same_tors = 0;
                        break;

               case 8:  sscanf(c, "%f %f", &tors_RB[0][i], &tors_RB[1][i]);
                        set_same_tors = 0;
                        break;

               default: sprintf(mess,
                          "Illegal torsion type in parameters, force field file %s\n",
                                myfiles);
                        my_fatal(FARGS, mess);
                        exit(1);
          }
          i++;
     }

/* get improper dihedrals parameters */
     countin = 0;
     while(fgets(line, MAXCHAR, ffptr) != NULL) {
          if(!strncmp(line, "impropers", 9)) {
               sscanf(&line[10], "%d", &countin);
               break;
          }
     }

     if(!countin) {
          sprintf(mess,
                  "Force field file %s lacks impropers count data\n",
                  myfiles);
          my_fatal(FARGS, mess);
          exit(1);
     }

     impro_name1 = (char **)mat_alloc(countin, 6, FARGS);
     impro_name2 = (char **)mat_alloc(countin, 6, FARGS);
     impro_name3 = (char **)mat_alloc(countin, 6, FARGS);
     impro_name4 = (char **)mat_alloc(countin, 6, FARGS);
     impro_mult = (int *)allocator(countin, sizeof(float), FARGS);
     impro_force = (float *)allocator(countin, sizeof(float), FARGS);
     impro_phase = (float *)allocator(countin, sizeof(float), FARGS);
     impro_term = (float *)allocator(countin, sizeof(float), FARGS);
     impro_xcount = (int *)allocator(countin, sizeof(int), FARGS);
     *nmimpro = countin;

     i = 0;
/* use of while instead of for to permit comments and blank lines */
     while(i < countin) {
          if(fgets(line, MAXCHAR, ffptr) == NULL) {
               sprintf(mess,
                       "Premature end to force field file %s\n",
                       myfiles);
               my_fatal(FARGS, mess);
               exit(1);
          }
          if((c = strchr(line, '\n')) != NULL)
               *c = NULLCHAR;                   /* strip returns */
          if((c = strchr(line, ';')) != NULL)
               *c = NULLCHAR;
          if((j = strspn(line, " \t")) == strlen(line))
               continue;                        /* drop blank lines */
          if(!strlen(line))
               continue;              /* drop empty strings as well */
          impro_force[i] = -1.0;
          impro_phase[i] = 99999.0;
          impro_term[i] = 1.0;
          c = get_name(line, impro_name3[i], 0);
          c = get_name(c, impro_name1[i], 0);
          c = get_name(c, impro_name2[i], 0);
          c = get_name(c, impro_name4[i], 0);
          impro_xcount[i] = 0;
          if(impro_name1[i][0] == 'X') impro_xcount[i]++;
          if(impro_name2[i][0] == 'X') impro_xcount[i]++;
          if(impro_name3[i][0] == 'X') impro_xcount[i]++;
          if(impro_name4[i][0] == 'X') impro_xcount[i]++;
          sscanf(c, "%f %f", &impro_phase[i], &impro_force[i]);
          impro_mult[i] = 1;
          if(impro_force[i] < 0.0 ) impro_mult[i] = 9999;
          i++;
     }

/* get names of aromatic C's that retain protons */
     countin = 0;
     while(fgets(line, MAXCHAR, ffptr) != NULL) {
          if(!strncmp(line, "arom", 4)) {
               sscanf(&line[5], "%d", &countin);
               break;
          }
     }

     if(!countin) {
          sprintf(mess,
                  "Force field file %s lacks aromatic C's count data\n",
                  myfiles);
          my_fatal(FARGS, mess);
          exit(1);
     }

     gmx_aroms = (char ***)tabl_alloc(countin, 2, 6, FARGS);
     *nmarom = countin;

     i = 0;
     while(i < countin) {
          if(fgets(line, MAXCHAR, ffptr) == NULL) {
               sprintf(mess,
                       "Premature end to force field file %s\n",
                       myfiles);
               my_fatal(FARGS, mess);
               exit(1);
          }
          if((c = strchr(line, '\n')) != NULL)
               *c = NULLCHAR;                   /* strip returns */
          if((c = strchr(line, ';')) != NULL)
               *c = NULLCHAR;
          if((j = strspn(line, " \t")) == strlen(line))
               continue;                        /* drop blank lines */
          if(!strlen(line))
               continue;              /* drop empty strings as well */
          c = get_name(line, gmx_aroms[i][0], 0);
          c = get_name(c, gmx_aroms[i][1], 0);
          i++;
     }

     fclose(ffptr);

     return;
}

/* Set torsion parameters from gromacs force field read in.
   Includes checks for residue specific torsion entries.
   parameters are:
        number of torsions in molecule:              num_tors
        number of torsion types in force field:      tors_cntFF
*/
void set_gromacs_torsFF(int num_tors, int tors_cntFF)
{
     int i, tors1, tors2, tors3, tors4, j, k;

     for(i = 0; i < num_tors; i++) {
          tors1 = torsion_tab[i][0];
          tors2 = torsion_tab[i][1];
          tors3 = torsion_tab[i][2];
          tors4 = torsion_tab[i][3];

          j = comp_gmx_tors(tors1, tors2, tors3, tors4,
                            tors_cntFF);

          if(j) a_tors_type[i] = FF_tors_type[(j - 1)];
          else a_tors_type[i] = 0;

          switch(a_tors_type[i]) {
               case 0: break;          /* torsion type not set. */

               case 1: a_tors_force[i] = tors_force[(j - 1)];
                       a_tors_mult[i] = tors_mult[(j - 1)];
                       a_tors_phase[i] = tors_phase[(j - 1)];
                       a_tors_term[i] = tors_term[(j - 1)];
                       if(j >= (tors_cntFF - 1))
                             break;
/* if this has multiple torsions, add links and set them as well
   note:  j points to the next torsion entry which is the next
   component of a multiple torsion and these facts are used below.
*/
                       if(tors_more[(j - 1)]) {
                            multors_cnt[i] = tors_more[(j - 1)];     /* set count */
                            multiples[i] = (float **)allocator(multors_cnt[i],
                                                               sizeof(float **),
                                                               FARGS);
                            for(k = 0; k < multors_cnt[i]; k++) {
                                 multiples[i][k] = (float *)allocator(4,
                                                                      sizeof(float),
                                                                      FARGS);
                                 multiples[i][k][0] = (float)tors_mult[j];
                                 multiples[i][k][1] = tors_force[j];
                                 multiples[i][k][2] = tors_phase[j];
                                 multiples[i][k][3] = tors_term[j];
                                 j++;
                            }
                       }
                       break;

               case 3: for(k = 0; k < 6; k++)
                            a_tors_set[k][i] = tors_RB[k][(j - 1)];
                       break;

               case 5: for(k = 0; k < 4; k++)
                            a_tors_set[k][i] = tors_RB[k][(j - 1)];
                       break;

               case 8: a_tors_set[0][i] = tors_RB[0][(j - 1)];
                       a_tors_set[1][i] = tors_RB[1][(j - 1)];
                       break;

               default: my_fatal(FARGS, "Illegal torsion type returned\n");
                        exit(1);
          }

     }

    return;
}

/* comparison of torsion to force field
   parameters are:
        atom types of each atom in torsion:               tors_atom1
                                                          tors_atom2
                                                          tors_atom3
                                                          tors_atom4
        number of torsion parameter sets in force field:  tors_cntFF
*/
int comp_gmx_tors(int tors_atom1, int tors_atom2, int tors_atom3,
                  int tors_atom4, int tors_cntFF)
{
     int j, k, l, no;

/* first try without wild cards */
     for(j = 0; j < tors_cntFF; j++) {
          no = 1;                            /* flag for special residue requirements */
/* check for dihedrals with special residue requirements.
   Uses the molecule's original residue names and not
   changed residue name.
*/
          while(gmx_tors_num[j]) {        /* use while instead of if for escapes below */
               if(strcmp(atom_residue[tors_atom1],
                         atom_residue[tors_atom2]) ||
                  strcmp(atom_residue[tors_atom1],
                         atom_residue[tors_atom3]) ||
                  strcmp(atom_residue[tors_atom1],
                         atom_residue[tors_atom4]))
                    break;           /* all atoms must be in the same original residue */
               l = 0;
               for(k = 0; k < gmx_tors_num[j]; k++)
                    if(!strcmp(atom_residue[tors_atom1],
                               gmx_tors_resid[j][k])) {
                         l = 1;
                         break;
                    }
/* if failed to find a needed match.  skip this force field entry */
               if(!l) no = 0;
               break;             /* really the while was an if, so need a break here */
          }
          if(!no) continue;              /* special residue requirements not met */

          if((!atom_cmp(a_ff_type[tors_atom2], tors_name2[j]) &&
              !atom_cmp(a_ff_type[tors_atom3], tors_name3[j])))    /* two center must match */
                    if((!atom_cmp(a_ff_type[tors_atom1],
                                   tors_name1[j])) &&        /* 1 goes to 2 and 4 goes to 3 */
                       (!atom_cmp(a_ff_type[tors_atom4],
                                  tors_name4[j])))          /* order here is important */
                              return(j + 1);

          if((!atom_cmp(a_ff_type[tors_atom2], tors_name3[j]) &&
              !atom_cmp(a_ff_type[tors_atom3], tors_name2[j])))   /* two center must match */
                    if((!atom_cmp(a_ff_type[tors_atom1],
                                  tors_name4[j])) &&       /* 4 goes to 2 and 1 goes to 3 */
                       (!atom_cmp(a_ff_type[tors_atom4],
                                  tors_name1[j])))             /* order here is important */
                              return(j + 1);
     }

/* then try with wild cards */
     for(j = 0; j < tors_cntFF; j++) {
          no = 1;                            /* flag for special residue requirements */
/* check for dihedrals with special residue requirements.
   Uses the molecule's original residue names and not
   changed residue name.
*/
          while(gmx_tors_num[j]) {        /* use while instead of if for escapes below */
               if(strcmp(atom_residue[tors_atom1],
                         atom_residue[tors_atom2]) ||
                  strcmp(atom_residue[tors_atom1],
                         atom_residue[tors_atom3]) ||
                  strcmp(atom_residue[tors_atom1],
                         atom_residue[tors_atom4]))
                    break;           /* all atoms must be in the same original residue */
               l = 0;
               for(k = 0; k < gmx_tors_num[j]; k++)
                    if(!strcmp(atom_residue[tors_atom1],
                               gmx_tors_resid[j][k])) {
                         l = 1;
                         break;
                    }
/* if failed to find a needed match.  skip this force field entry */
               if(!l) no = 0;
               break;             /* really the while was an if, so need a break here */
          }
          if(!no) continue;              /* special residue requirements not met */

          if((!atom_cmp(a_ff_type[tors_atom2], tors_name2[j]) &&
              !atom_cmp(a_ff_type[tors_atom3], tors_name3[j])))    /* two center must match */
                    if(((!atom_cmp(a_ff_type[tors_atom1], tors_name1[j])) ||
                        (tors_name1[j][0] == 'X')) &&      /* 1 goes to 2 and 4 goes to 3 */
                       ((!atom_cmp(a_ff_type[tors_atom4], tors_name4[j])) ||
                        (tors_name4[j][0] == 'X')))        /* order here is important */
                              return(j + 1);

          if((!atom_cmp(a_ff_type[tors_atom2], tors_name3[j]) &&
              !atom_cmp(a_ff_type[tors_atom3], tors_name2[j])))   /* two center must match */
                    if(((!atom_cmp(a_ff_type[tors_atom1], tors_name4[j])) ||
                        (tors_name4[j][0] == 'X')) &&     /* 4 goes to 2 and 1 goes to 3 */
                       ((!atom_cmp(a_ff_type[tors_atom4], tors_name1[j])) ||
                        (tors_name1[j][0] == 'X')))       /* order here is important */
                              return(j + 1);
     }

     return(0);                                            /* nothing matched */
}
